@@ -58,9 +58,42 @@ def lineno_from_content_start(content: str, start: int) -> int:
     return content[:start].count('\n') + 1
 
 
+def safe_os_path_relpath(path: str, start: str) -> str:
+    """Return the relative path of a file from a start directory.
+
+    Safe version of `os.path.relpath` that catches `ValueError` exceptions
+    on Windows and returns the original path in case of error.
+    On Windows, `ValueError` is raised when `path` and `start` are on
+    different drives.
+    """
+    if os.name != 'nt':  # pragma: nt no cover
+        return os.path.relpath(path, start)
+    try:  # pragma: nt cover
+        return os.path.relpath(path, start)
+    except ValueError:  # pragma: no cover
+        return path
+
+
+def file_lineno_message(
+        page_src_path: str | None,
+        docs_dir: str,
+        lineno: int,
+) -> str:
+    """Return a message with the file path and line number."""
+    if page_src_path is None:  # pragma: no cover
+        return f'generated page content (line {lineno})'
+    return (
+        f'{safe_os_path_relpath(page_src_path, docs_dir)}'
+        f':{lineno}'
+    )
+
+
 def get_file_content(  # noqa: PLR0913, PLR0915
         markdown: str,
-        page_src_path: str,
+        # Generated pages return `None` for `file.abs_src_path` because
+        # they are not read from a file. In this case, page_src_path is
+        # set to `None`.
+        page_src_path: str | None,
         docs_dir: str,
         tags: IncludeTags,
         defaults: DefaultValues,
@@ -70,6 +103,22 @@ def get_file_content(  # noqa: PLR0913, PLR0915
         http_cache: Cache | None = None,
 ) -> str:
     """Return the content of the file to include."""
+    settings_ignore_paths = []
+    if settings.exclude is not None:
+        for path in glob.glob(
+                [
+                    os.path.join(docs_dir, fp)
+                    if not os.path.isabs(fp)
+                    else fp for fp in settings.exclude
+                ],
+                flags=GLOB_FLAGS,
+                root_dir=docs_dir,
+        ):
+            if path not in settings_ignore_paths:
+                settings_ignore_paths.append(path)
+        if page_src_path in settings_ignore_paths:
+            return markdown
+
     def found_include_tag(  # noqa: PLR0912, PLR0915
             match: re.Match[str],
     ) -> str:
@@ -85,26 +134,14 @@ def get_file_content(  # noqa: PLR0913, PLR0915
             )
             raise PluginError(
                 "Found no path passed including with 'include'"
-                f' directive at {os.path.relpath(page_src_path, docs_dir)}'
-                f':{lineno}',
+                ' directive at'
+                f' {file_lineno_message(page_src_path, docs_dir, lineno)}',
             )
 
         arguments_string = match['arguments']
 
         exclude_match = ARGUMENT_REGEXES['exclude'].search(arguments_string)
-        ignore_paths = []
-        if settings.exclude:
-            ignore_paths.extend(
-                glob.glob(
-                    [
-                        os.path.join(docs_dir, fp)
-                        if not os.path.isabs(fp)
-                        else fp for fp in settings.exclude
-                    ],
-                    flags=GLOB_FLAGS,
-                    root_dir=docs_dir,
-                ),
-            )
+        ignore_paths = [*settings_ignore_paths]
         if exclude_match is not None:
             exclude_string = parse_string_argument(exclude_match)
             if exclude_string is None:
@@ -114,17 +151,17 @@ def get_file_content(  # noqa: PLR0913, PLR0915
                 )
                 raise PluginError(
                     "Invalid empty 'exclude' argument in 'include'"
-                    f' directive at {os.path.relpath(page_src_path, docs_dir)}'
-                    f':{lineno}',
+                    ' directive at'
+                    f' {file_lineno_message(page_src_path, docs_dir, lineno)}',
                 )
-            ignore_paths.extend(
-                resolve_file_paths_to_exclude(
+
+            for path in resolve_file_paths_to_exclude(
                     exclude_string,
                     page_src_path,
                     docs_dir,
-                ),
-            )
-        ignore_paths = list(set(ignore_paths))
+            ):
+                if path not in ignore_paths:
+                    ignore_paths.append(path)
 
         file_paths_to_include, is_url = resolve_file_paths_to_include(
             filename,
@@ -139,16 +176,16 @@ def get_file_content(  # noqa: PLR0913, PLR0915
                 directive_match_start,
             )
             raise PluginError(
-                f"No files found including '{raw_filename}'"
-                f' at {os.path.relpath(page_src_path, docs_dir)}'
-                f':{lineno}',
+                f"No files found including '{raw_filename}' at"
+                f' {file_lineno_message(page_src_path, docs_dir, lineno)}',
             )
 
         if files_watcher is not None and not is_url:
             files_watcher.included_files.extend(file_paths_to_include)
 
         bool_options, invalid_bool_args = parse_bool_options(
-            ['preserve-includer-indent', 'dedent', 'trailing-newlines'],
+            ['preserve-includer-indent', 'dedent',
+                'trailing-newlines', 'recursive'],
             defaults,
             arguments_string,
         )
@@ -160,8 +197,8 @@ def get_file_content(  # noqa: PLR0913, PLR0915
             raise PluginError(
                 f"Invalid value for '{invalid_bool_args[0]}' argument of"
                 " 'include' directive at"
-                f' {os.path.relpath(page_src_path, docs_dir)}'
-                f':{lineno}. Possible values are true or false.',
+                f' {file_lineno_message(page_src_path, docs_dir, lineno)}.'
+                f' Possible values are true or false.',
             )
 
         start_match = ARGUMENT_REGEXES['start'].search(arguments_string)
@@ -173,8 +210,8 @@ def get_file_content(  # noqa: PLR0913, PLR0915
                     directive_match_start,
                 )
                 raise PluginError(
-                    "Invalid empty 'start' argument in 'include' directive at "
-                    f'{os.path.relpath(page_src_path, docs_dir)}:{lineno}',
+                    "Invalid empty 'start' argument in 'include' directive at"
+                    f' {file_lineno_message(page_src_path, docs_dir, lineno)}',
                 )
         else:
             start = defaults['start']
@@ -188,8 +225,8 @@ def get_file_content(  # noqa: PLR0913, PLR0915
                     directive_match_start,
                 )
                 raise PluginError(
-                    "Invalid empty 'end' argument in 'include' directive at "
-                    f'{os.path.relpath(page_src_path, docs_dir)}:{lineno}',
+                    "Invalid empty 'end' argument in 'include' directive at"
+                    f' {file_lineno_message(page_src_path, docs_dir, lineno)}',
                 )
         else:
             end = defaults['end']
@@ -204,8 +241,8 @@ def get_file_content(  # noqa: PLR0913, PLR0915
                 )
                 raise PluginError(
                     "Invalid empty 'encoding' argument in 'include'"
-                    ' directive at '
-                    f'{os.path.relpath(page_src_path, docs_dir)}:{lineno}',
+                    ' directive at'
+                    f' {file_lineno_message(page_src_path, docs_dir, lineno)}',
                 )
         else:
             encoding = defaults['encoding']
@@ -231,16 +268,17 @@ def get_file_content(  # noqa: PLR0913, PLR0915
                         expected_but_any_found[i] = False
 
             # nested includes
-            new_text_to_include = get_file_content(
-                new_text_to_include,
-                file_path,
-                docs_dir,
-                tags,
-                defaults,
-                settings,
-                files_watcher=files_watcher,
-                http_cache=http_cache,
-            )
+            if bool_options['recursive'].value:
+                new_text_to_include = get_file_content(
+                    new_text_to_include,
+                    file_path,
+                    docs_dir,
+                    tags,
+                    defaults,
+                    settings,
+                    files_watcher=files_watcher,
+                    http_cache=http_cache,
+                )
 
             # trailing newlines right stripping
             if not bool_options['trailing-newlines'].value:
@@ -270,7 +308,7 @@ def get_file_content(  # noqa: PLR0913, PLR0915
             if expected_but_any_found[i]:
                 delimiter_value = locals()[delimiter_name]
                 readable_files_to_include = ', '.join([
-                    os.path.relpath(fpath, docs_dir)
+                    safe_os_path_relpath(fpath, docs_dir)
                     for fpath in file_paths_to_include
                 ])
                 plural_suffix = 's' if len(file_paths_to_include) > 1 else ''
@@ -278,12 +316,14 @@ def get_file_content(  # noqa: PLR0913, PLR0915
                     markdown,
                     directive_match_start,
                 )
-                relative_path = os.path.relpath(page_src_path, docs_dir)
+                file_lineno = file_lineno_message(
+                    page_src_path, docs_dir, lineno,
+                )
                 logger.warning(
                     (
                         f"Delimiter {delimiter_name} '{delimiter_value}'"
                         " of 'include' directive at"
-                        f' {relative_path}:{lineno}'
+                        f' {file_lineno}'
                         f' not detected in the file{plural_suffix}'
                         f' {readable_files_to_include}'
                     ),
@@ -307,26 +347,14 @@ def get_file_content(  # noqa: PLR0913, PLR0915
             )
             raise PluginError(
                 "Found no path passed including with 'include-markdown'"
-                f' directive at {os.path.relpath(page_src_path, docs_dir)}'
-                f':{lineno}',
+                f' directive at'
+                f' {file_lineno_message(page_src_path, docs_dir, lineno)}',
             )
 
         arguments_string = match['arguments']
 
         exclude_match = ARGUMENT_REGEXES['exclude'].search(arguments_string)
-        ignore_paths = []
-        if settings.exclude is not None:
-            ignore_paths.extend(
-                glob.glob(
-                    [
-                        os.path.join(docs_dir, fp)
-                        if not os.path.isabs(fp)
-                        else fp for fp in settings.exclude
-                    ],
-                    flags=GLOB_FLAGS,
-                    root_dir=docs_dir,
-                ),
-            )
+        ignore_paths = [*settings_ignore_paths]
         if exclude_match is not None:
             exclude_string = parse_string_argument(exclude_match)
             if exclude_string is None:
@@ -336,17 +364,16 @@ def get_file_content(  # noqa: PLR0913, PLR0915
                 )
                 raise PluginError(
                     "Invalid empty 'exclude' argument in 'include-markdown'"
-                    f' directive at {os.path.relpath(page_src_path, docs_dir)}'
-                    f':{lineno}',
+                    f' directive at'
+                    f' {file_lineno_message(page_src_path, docs_dir, lineno)}',
                 )
-            ignore_paths.extend(
-                resolve_file_paths_to_exclude(
+            for path in resolve_file_paths_to_exclude(
                     exclude_string,
                     page_src_path,
                     docs_dir,
-                ),
-            )
-        ignore_paths = list(set(ignore_paths))
+            ):
+                if path not in ignore_paths:
+                    ignore_paths.append(path)
 
         file_paths_to_include, is_url = resolve_file_paths_to_include(
             filename,
@@ -362,8 +389,7 @@ def get_file_content(  # noqa: PLR0913, PLR0915
             )
             raise PluginError(
                 f"No files found including '{raw_filename}' at"
-                f' {os.path.relpath(page_src_path, docs_dir)}'
-                f':{lineno}',
+                f' {file_lineno_message(page_src_path, docs_dir, lineno)}',
             )
 
         if files_watcher is not None and not is_url:
@@ -386,8 +412,8 @@ def get_file_content(  # noqa: PLR0913, PLR0915
             raise PluginError(
                 f"Invalid value for '{invalid_bool_args[0]}' argument of"
                 " 'include-markdown' directive at"
-                f' {os.path.relpath(page_src_path, docs_dir)}'
-                f':{lineno}. Possible values are true or false.',
+                f' {file_lineno_message(page_src_path, docs_dir, lineno)}.'
+                f' Possible values are true or false.',
             )
 
         # start and end arguments
@@ -401,8 +427,8 @@ def get_file_content(  # noqa: PLR0913, PLR0915
                 )
                 raise PluginError(
                     "Invalid empty 'start' argument in 'include-markdown'"
-                    f' directive at {os.path.relpath(page_src_path, docs_dir)}'
-                    f':{lineno}',
+                    f' directive at'
+                    f' {file_lineno_message(page_src_path, docs_dir, lineno)}',
                 )
         else:
             start = defaults['start']
@@ -417,8 +443,8 @@ def get_file_content(  # noqa: PLR0913, PLR0915
                 )
                 raise PluginError(
                     "Invalid empty 'end' argument in 'include-markdown'"
-                    f' directive at {os.path.relpath(page_src_path, docs_dir)}'
-                    f':{lineno}',
+                    f' directive at'
+                    f' {file_lineno_message(page_src_path, docs_dir, lineno)}',
                 )
         else:
             end = defaults['end']
@@ -433,8 +459,8 @@ def get_file_content(  # noqa: PLR0913, PLR0915
                 )
                 raise PluginError(
                     "Invalid empty 'encoding' argument in 'include-markdown'"
-                    ' directive at '
-                    f'{os.path.relpath(page_src_path, docs_dir)}:{lineno}',
+                    ' directive at'
+                    f' {file_lineno_message(page_src_path, docs_dir, lineno)}',
                 )
         else:
             encoding = defaults['encoding']
@@ -444,8 +470,8 @@ def get_file_content(  # noqa: PLR0913, PLR0915
             arguments_string,
         )
         if offset_match:
-            offset = offset_match[1]
-            if offset == '':
+            offset_raw_value = offset_match[1]
+            if offset_raw_value == '':
                 lineno = lineno_from_content_start(
                     markdown,
                     directive_match_start,
@@ -453,19 +479,19 @@ def get_file_content(  # noqa: PLR0913, PLR0915
                 raise PluginError(
                     "Invalid empty 'heading-offset' argument in"
                     " 'include-markdown' directive at"
-                    f' {os.path.relpath(page_src_path, docs_dir)}:{lineno}',
+                    f' {file_lineno_message(page_src_path, docs_dir, lineno)}',
                 )
             try:
-                offset = int(offset)
+                offset = int(offset_raw_value)
             except ValueError:
                 lineno = lineno_from_content_start(
                     markdown,
                     directive_match_start,
                 )
                 raise PluginError(
-                    f"Invalid 'heading-offset' argument \"{offset}\" in"
-                    " 'include-markdown' directive at "
-                    f'{os.path.relpath(page_src_path, docs_dir)}:{lineno}',
+                    f"Invalid 'heading-offset' argument \"{offset_raw_value}\""
+                    " in 'include-markdown' directive at"
+                    f' {file_lineno_message(page_src_path, docs_dir, lineno)}',
                 ) from None
         else:
             offset = defaults['heading-offset']
@@ -523,11 +549,17 @@ def get_file_content(  # noqa: PLR0913, PLR0915
 
             # relative URLs rewriting
             if bool_options['rewrite-relative-urls'].value:
-                new_text_to_include = process.rewrite_relative_urls(
-                    new_text_to_include,
-                    source_path=file_path,
-                    destination_path=page_src_path,
-                )
+                if page_src_path is None:  # pragma: no cover
+                    logger.warning(
+                        'Relative URLs rewriting is not supported in'
+                        ' generated pages.',
+                    )
+                else:
+                    new_text_to_include = process.rewrite_relative_urls(
+                        new_text_to_include,
+                        source_path=file_path,
+                        destination_path=page_src_path,
+                    )
 
             # comments
             if bool_options['comments'].value:
@@ -556,7 +588,7 @@ def get_file_content(  # noqa: PLR0913, PLR0915
                     )
                 )
 
-            if offset_match:
+            if offset:
                 new_text_to_include = process.increase_headings_offset(
                     new_text_to_include,
                     offset=offset + cumulative_heading_offset,
@@ -569,7 +601,7 @@ def get_file_content(  # noqa: PLR0913, PLR0915
             if expected_but_any_found[i]:
                 delimiter_value = locals()[delimiter_name]
                 readable_files_to_include = ', '.join([
-                    os.path.relpath(fpath, docs_dir)
+                    safe_os_path_relpath(fpath, docs_dir)
                     for fpath in file_paths_to_include
                 ])
                 plural_suffix = 's' if len(file_paths_to_include) > 1 else ''
@@ -577,12 +609,14 @@ def get_file_content(  # noqa: PLR0913, PLR0915
                     markdown,
                     directive_match_start,
                 )
-                relative_path = os.path.relpath(page_src_path, docs_dir)
+                file_lineno = file_lineno_message(
+                    page_src_path, docs_dir, lineno,
+                )
                 logger.warning(
                     (
                         f"Delimiter {delimiter_name} '{delimiter_value}' of"
                         " 'include-markdown' directive at"
-                        f' {relative_path}:{lineno}'
+                        f' {file_lineno}'
                         f' not detected in the file{plural_suffix}'
                         f' {readable_files_to_include}'
                     ),
@@ -625,6 +659,7 @@ def on_page_markdown(
             'comments': config.comments,
             'rewrite-relative-urls': config.rewrite_relative_urls,
             'heading-offset': config.heading_offset,
+            'recursive': config.recursive,
             'start': config.start,
             'end': config.end,
         },
